import os

from .paths import PathsConf
from ..descriptors import env

# Map country on default configurations.
COUNTRY_DB = {
    None: {
        'LANGUAGE_CODE': 'en-us',
        'TIME_ZONE': 'Greenwich',
    },
}
COUNTRY_DB[''] = COUNTRY_DB[None]


class LocaleConf(PathsConf):
    """
    Base Django locale configuration.
    """
    USE_I18N = env(True)
    USE_L10N = env(True)
    USE_TZ = env(True)
    COUNTRY = env('', name='{attr}')

    def get_language_code(self, country):
        return country_method('language_code', country)

    def get_time_zone(self, country):
        return country_method('time_zone', country)

    def get_locale_name(self, language_code, country):
        try:
            return country_method('locale_name', country)
        except ValueError:
            if language_code:
                lang, country = language_code.split('-')
                return f'{lang}_{country.upper()}.UTF8'
            else:
                return ''

    def get_locale_paths(self):
        path = os.path.join(self.REPO_DIR, 'locale')
        return self.env('LOCALE_PATH', type=list, default=[path])


class NoLocaleConf(PathsConf):
    """
    Base configuration for instances that do not use localization.
    """
    USE_I18N = env(False)
    USE_L10N = env(False)
    USE_TZ = False
    LANGUAGE_CODE = None
    TIME_ZONE = None


def country_method(var, country):
    django_var = 'DJANGO_' + var.upper()
    key = var.upper()
    if django_var in os.environ:
        return os.environ[django_var]
    try:
        return COUNTRY_DB[country.upper()][key]
    except KeyError as exc:
        raise ValueError(f'{exc}: unknown for country {country}')


#
# Register countries
#
def country(country: list, language_code, timezone):
    """
    Return a function that creates a LocaleConf subclass specialized for a
    given country.

    This function return a factory rather than the Conf subclass itself to avoid
    the class creation overhead of a big list of countries.
    """
    names = [country] if isinstance(country, str) else country
    for alias in names:
        COUNTRY_DB[alias.upper()] = {
            'LANGUAGE_CODE': language_code,
            'TIME_ZONE': timezone,
        }


# Ursal
country(['Argentina'], 'es-ar', 'America/Buenos_Aires')
country(['Bolívia'], 'es-bo', 'America/La_Paz')
country(['Brasil', 'Brazil'], 'pt-br', 'America/Sao_Paulo')
country(['Chile'], 'es-cl', 'America/Santiago')
country(['Ecuador'], 'es-ec', 'America/Guayaquil')
country(['Paraguay'], 'es-py', 'America/Asuncion')
country(['Peru'], 'es-pe', 'America/Lima')
country(['Uruguay'], 'es-uy', 'America/Montevideo')
country(['Venezuela'], 'es-ve', 'America/Caracas')

# Other american countries

# Africa

# Asia

# Europe

# Oceania



import re
import os
from setuptools import setup, find_packages

init = open(os.path.join('src', 'boogie', '__init__.py')).read()
version = re.search(r"__version__ = '(\d+\.\d+.\d+)'", init).group(1)

setup(
    name='django-boogie',
    version=version,
    package_dir={'': 'src'},
    packages=find_packages('src'),
    setup_requires='setuptools >= 30.3',
)



from pprint import pprint

from sidekick import fn

from .utils.text import safe_repr, first_line


#
# Printing
#
@fn
def info(obj, log=print, full=False):
    """
    Display detailed information about object.
    """

    log('=' * 60)
    log('Object:', obj)

    # Print object state
    try:
        dic = obj.__dict__
    except (AttributeError, TypeError):
        pass
    else:
        log('\nDict')
        if full:
            pprint(dic)
        else:
            _log_values(log, dic)

    # Print information about methods
    methods = []
    for attr, method in public_methods(obj):
        doc = method.__doc__ or ''
        if doc:
            methods.append(f'* {attr}: {first_line(doc)}')
        else:
            methods.append(f'* {attr}')
    if methods:
        log('\nMethods:')
        print('\n'.join(methods))
    log('-' * 60)


def _log_values(log, values):
    # First print public values
    for k, v in values.items():
        if not k.startswith('_'):
            log(f'* {k}:', safe_repr(v, max_length=50))

    # Then the private ones (only types)
    for k, v in values.items():
        if k.startswith('_'):
            log(f'* {k} ({v.__class__.__name__})')


#
# Tracing
#
def set_trace():
    raise NotImplementedError


def embed():
    raise NotImplementedError


#
# Integration
#
def namespace():
    """
    Return a namespace of useful debug functions.
    """
    import boogie.debug as dbg

    return {
        'info': info, 'log': print, 'pprint': pprint, 'set_trace': set_trace,
        'embed': embed, 'dbg': dbg,
    }


def enable_debugging(auto=False):
    """
    Register debugging function to builtins.
    """
    import builtins

    # Checks if debugging should be enabled.
    if auto:
        pass

    # Configure environment in "debug" mode.
    for k, v in namespace().items():
        if not hasattr(builtins, k):
            setattr(builtins, k, v)


#
# Utility
#
def public_dir(obj):
    return (attr for attr in dir(obj) if not attr.startswith('_'))


def public_vars(obj):
    for attr in public_dir(obj):
        try:
            yield attr, getattr(obj, attr)
        except AttributeError:
            pass


def public_methods(obj):
    for attr, method in public_vars(obj):
        if callable(method):
            yield attr, method



from .environment import EnvironmentConf
from ..descriptors import env_settings


def default(value):
    return env_settings(default=value)(lambda self, env: env)


class UrlsConf(EnvironmentConf):
    """
    Configure urls for your Django project.
    """

    get_admin_url = default('/admin/')
    get_login_url = default('/login/')
    get_logout_url = default('/logout/')
    get_static_url = default('/static/')
    get_media_url = default('/media/')

    def get_append_slash(self):
        return True

    @env_settings(default=None)
    def get_root_urlconf(self, env):
        return env or self.get_django_project_path() + '.urls'



import inspect

import factory as _factory
from factory import DjangoModelFactory, Factory
from factory import declarations
from factory.declarations import BaseDeclaration
from faker import Factory as FakeFactory
from model_mommy.mommy import Mommy

fake = FakeFactory.create()

FactoryMeta = type(DjangoModelFactory)


def __getitem__(self, idx):
    if isinstance(idx, slice):
        pass
    return ...


FactoryMeta.__getitem__ = __getitem__


class BoogieMommy(Mommy):
    """
    Base for Boogie Model Mommy classes.
    """
    attr_mapping = {}
    type_mapping = {}


#
# Declarations
#
class ImplicitDeclaration(BaseDeclaration):
    def __init__(self):
        pass


#
# Factory function
#
def factory(model, **kwargs):
    """
    Creates a factory boy factory class
    """

    ns = {'Meta': type('Meta', (), {'model': model})}

    # Handle explicitly declared values
    for k, v in kwargs:
        ns[k] = explicit_declaration(model, k, v)

    # Create mommy instance to help with automatic value generation
    mommy = Mommy(model)

    # Create implicit declarations
    for field in model._meta.fields:
        if not requires_declaration(model, field.name, ns):
            continue
        ns[field.name] = implicit_declaration(model, field.name, ns, mommy)

    return type(model.__name__ + 'Factory', (DjangoModelFactory,), ns)


def explicit_declaration(model, name, value):
    """
    Return a Declaration instance that implements an explicitly defined field
    for a model.

    Args:
        model:
            Model class
        name:
            Name of the field in the model
        value:
            Value explicitly passed by the user
    """
    if issubclass(value, BaseDeclaration):
        return value
    elif isinstance(value, type) and issubclass(value, Factory):
        return _factory.SubFactory(value)
    elif callable(value):
        if has_no_args(value):
            return _factory.LazyFunction(value)
        else:
            return _factory.LazyAttribute(value)
    elif isinstance(value, str):
        return _factory.LazyAttribute(lambda x: value.format(model=x))
    else:
        return value


def requires_declaration(model, name, definitions):
    """
    Return True if explicit generation of given field is required during model
    instantiation.

    Args:
        model:
            Model class
        name:
            Model field name
        definitions:
            A map of names of all explicitly defined fields to their
            corresponding defined values.
    """
    field = model._meta.get_field(name)
    if field.has_default() or field.auto_created:
        return False
    if field in definitions:
        return False
    return True


def implicit_declaration(model, name, definitions, mommy):
    """
    Creates an implicit declaration for the field.

    Receives the same arguments as :func:`requires_declaration`, but returns
    a declaration instance.
    """

    field = model._meta.get_field(name)
    try:
        faker = getattr(fake, name)
        return declarations.LazyFunction(faker)
    except AttributeError:
        generator = (lambda: mommy.generate_value(field, commit=False))
        return declarations.LazyFunction(generator)


def has_no_args(func):
    """
    Return True if function is called with no positional args.
    """
    try:
        spec = inspect.getfullargspec(func)
    except TypeError:
        return has_no_args(func.__call__)
    return bool(spec.args)



from django.db import models
from sidekick import delegate_to

from .methodregistry import get_manager_attribute
from .queryset import QuerySet


class Manager(models.Manager):
    """
    Default Boogie manager.
    """

    # Delegate to queryset
    _queryset_class = QuerySet
    to_pivot_table = delegate_to('_queryset')
    to_timeseries = delegate_to('_queryset')
    to_dataframe = delegate_to('_queryset')
    id_dict = delegate_to('_queryset')
    bulk_upsert = delegate_to('_queryset')
    sync = delegate_to('_queryset')
    upsert = delegate_to('_queryset')
    get_or_none = delegate_to('_queryset')
    single = delegate_to('_queryset')
    bulk_update = delegate_to('_queryset')
    index = delegate_to('_queryset')
    dataframe = delegate_to('_queryset')
    select_columns = delegate_to('_queryset')
    update_item = delegate_to('_queryset')
    update_from_dataframe = delegate_to('_queryset')
    pivot_table = delegate_to('_queryset')
    extend_dataframe = delegate_to('_queryset')
    head = delegate_to('_queryset')
    tail = delegate_to('_queryset')

    def __getitem__(self, item):
        return self.get_queryset().__getitem__(item)

    def __setitem__(self, item, value):
        return self.get_queryset().__setitem__(item, value)

    def __getattr__(self, attr):
        return get_manager_attribute(self, attr)

    def get_queryset(self):
        return self._queryset_class(model=self.model, using=self._db, hints=self._hints)

    _queryset = property(get_queryset)

    def new(self, *args, **kwargs):
        """
        Create a new model instance, without saving it to the database.
        """
        return self.model(*args, **kwargs)



from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models.base import ModelBase as DjangoModelBase
from sidekick import import_later

from .manager import Manager
from .utils import with_base

model_utils = import_later('model_utils.models')
polymorphic = import_later('polymorphic.models')


class ModelBase(DjangoModelBase):
    """
    Base metaclass for Boogie models.
    """

    def __new__(metacls, name, bases, ns, **kwargs):
        # Extract the correct base class
        bases, kwargs = extract_bases(bases, **kwargs)

        # Is abstract
        if 'Meta' in ns:
            is_abstract = getattr(ns['Meta'], 'abstract', False)
        else:
            is_abstract = False

        # Add default manager to class
        if not is_abstract and 'objects' not in ns:
            ns['objects'] = manager = Manager()
            manager.auto_created = True

        # Additional arguments are passed to the Meta class object
        if kwargs and 'Meta' in ns:
            raise ImproperlyConfigured(
                'Cannot pass meta arguments to class constructor and the Meta '
                'class simultaneously.'
            )
        if kwargs:
            ns['Meta'] = type('Meta', (), kwargs)

        # Create class using default methods.
        return super().__new__(metacls, name, bases, ns)


#
# Utility
#
def extract_bases(bases, timestamped=False, timeframed=False, status=False,
                  soft_deletable=False, polymorphic=False, **kwargs):
    # Inject custom base classes from metaclass options.
    if timeframed:
        bases = with_base(bases, model_utils.TimeFramedModel)
    if timestamped:
        bases = with_base(bases, model_utils.TimeStampedModel)
    if status:
        bases = with_base(bases, model_utils.StatusModel)
    if soft_deletable:
        bases = with_base(bases, model_utils.SoftDeletableModel)
    if polymorphic:
        bases = with_base(bases, polymorphic.PolymorphicModel)

    return bases, kwargs


#
# Base model
#
class Model(models.Model, metaclass=ModelBase, abstract=True):
    """
    Base class for Boogie model.
    """



from django.apps import apps as _apps

# Import all fields and models. boogie.models serves as a drop-in replacement
# for django models.
from django.db.models import *

# Overload Django objects
from .manager import Manager
from .queryset import QuerySet
from .model import Model
from .expressions import F, concat, coalesce, greatest, least

# Load Boogie features
from .methodregistry import manager_method, queryset_method
from ..fields import *


from copy import copy

from django.db.models import Model, AutoField, FieldDoesNotExist
from django.urls import reverse
from rest_framework.serializers import SerializerMethodField
from sidekick import lazy

from .serializers import RestAPISerializer, RestAPIInlineSerializer
from .settings import get_url_prefix
from .utils import natural_base_url, viewset_actions
from .viewsets import RestAPIBaseViewSet


class ResourceInfo:
    """
    Stores all information about a resource that is necessary to build the
    corresponding serializer and viewset classes.
    """

    model_name = lazy(lambda self: self.model.__name__)

    @property
    def field_names(self):
        yield from (f.name for f in self.model._meta.fields)

    @property
    def related_field_names(self):
        yield from (name for name, model in self.related_models)

    @property
    def related_models(self):
        for name in self.fields:
            if name in self.properties:
                pass
            try:
                field = self.meta.get_field(name)
            except FieldDoesNotExist:
                continue
            if field.related_model:
                yield name, field.related_model

    @lazy
    def property_methods(self):
        ns = {}
        for name, property in self.properties.items():
            ns[name] = SerializerMethodField()
            ns['get_' + name] = property_method(property, name)
        return ns

    @lazy
    def queryset(self):
        """
        Default queryset for the resource.
        """
        fields = self.related_field_names
        qs = self.model._default_manager.select_related(*fields)
        return self.update_queryset(qs)

    @lazy
    def action_methods(self):
        """
        Methods registered with the @rest_api.action() decorator.
        """
        return viewset_actions(self.actions)

    @lazy
    def detail_actions(self):
        return {k: v for k, v in self.actions.items() if v['args'].get('detail')}

    @lazy
    def serializer_hook_methods(self):
        methods = {}

        if 'save' in self.hooks:
            save_hook = wrap_request_instance_method(self.hooks['save'])
            methods['save_hook'] = save_hook

        return methods

    @lazy
    def viewset_hook_methods(self):
        methods = {}
        for hook in ('delete', 'query'):
            if hook in self.hooks:
                hook_method = wrap_request_instance_method(self.hooks[hook])
                methods[hook + '_hook'] = hook_method

        return methods

    def __init__(self,
                 model: Model,

                 # Fields
                 fields=None, exclude=(),

                 # Urls and views
                 base_url=None, base_name=None,

                 # Viewset options
                 viewset_base=RestAPIBaseViewSet,
                 update_queryset=lambda x: x,

                 # Serializer options
                 serializer_base=None,

                 # Other options
                 inline=False, lookup_field='pk'):

        self.model = model
        self.meta = model._meta
        self._model_fields = {f.name: f for f in self.meta.fields}
        self.inline = inline
        self.lookup_field = lookup_field

        # Field info
        fields = list(fields or fields_from_model(model))
        for field in exclude:
            if field in fields:
                fields.remove(field)
        self.fields = []
        for f in fields:
            self.add_field(f)

        # Url info
        self.base_url = base_url or natural_base_url(model)
        self.base_name = base_name or natural_base_url(model)

        # Hooks
        self.actions = {}
        self.properties = {}
        self.hooks = {}
        self.links = {}

        # Viewsets
        self.viewset_base = viewset_base
        self.update_queryset = update_queryset

        # Serializer
        if serializer_base is None:
            if inline:
                serializer_base = RestAPIInlineSerializer
            else:
                serializer_base = RestAPISerializer
        self.serializer_base = serializer_base

    def copy(self):
        """
        Return a copy of itelf.
        """
        return copy(self)

    def add_hook(self, hook, function):
        if hook not in ('save', 'delete', 'query'):
            raise ValueError(f'invalid hook: {hook}')
        self.hooks[hook] = function

    def add_field(self, name, check=True):
        """
        Register a new field name.
        """
        if check:
            self.meta.get_field(name)
        if name not in self.fields:
            self.fields.append(name)

    def add_link(self, name, method):
        """
        Register a new field name.
        """
        self.links[name] = method

    def add_property(self, name, method):
        """
        Register a property with the provided name.

        Args:
            name:
                Property name.
            method:
                A function f(obj) -> value that receives a model instance and
                return the corresponding property value.
        """
        self.add_field(name, check=False)
        self.properties[name] = method

    def add_action(self, name, method, **kwargs):
        """
        Register an action with the given name.
        """
        if 'list' in kwargs and 'detail' in kwargs:
            msg = ('cannot specify both "list" and "detail" parameters '
                   'simultaneously')
            raise TypeError(msg)
        elif 'list' in kwargs:
            kwargs['detail'] = not kwargs.pop('list')
        kwargs.setdefault('detail', True)
        self.actions[name] = {'method': method, 'args': kwargs}

    #
    # Info
    #
    def full_base_name(self, version):
        """
        Base name with version information.
        """
        return '%s-%s' % (version, self.base_name)

    def extra_kwargs(self, version):
        """
        extra_kwargs applied on the serializer for references to the resource
        model.
        """
        return {
            'view_name': self.full_base_name(version) + '-detail',
            'lookup_field': self.lookup_field,
        }

    def detail_hyperlink(self, obj, request=None, version=None):
        """
        Return API hyperlink for object.
        """
        attr = self.lookup_field
        kwargs = {attr: getattr(obj, attr)}
        if version:
            view_name = self.full_base_name(version)
        else:
            view_name = self.base_name
        path = reverse(view_name + '-detail', kwargs=kwargs)
        return get_url_prefix(request) + path


#
# Auxiliary functions
#
def fields_from_model(model):
    return [f.name for f in model._meta.fields if not isinstance(f, AutoField)]


def property_method(func, name):
    def method(self, obj):
        return func(obj)

    method.__name__ = method.__qualname__ = 'get_' + name
    return method


def wrap_request_instance_method(func):
    return lambda self, request, obj: func(request, obj)



import importlib.util

from sidekick import unique
from .environment import EnvironmentConf


class InstalledAppsConf(EnvironmentConf):
    """
    Mixin class that helps managing INSTALLED_APPS.
    """

    def get_use_django_users(self):
        return True

    def get_use_django_admin(self):
        return True

    def get_installed_apps(self):
        """
        Return a list of installed apps.

        The list contains project  apps + thirdy parties + contrib, in that
        order.
        """
        return list(unique(filter(lambda x: x is not None, [
            *self.PROJECT_APPS,
            *self.THIRD_PARTY_APPS,
            *self.DJANGO_CONTRIB_APPS,
        ])))

    def get_project_apps(self):
        """
        Return a list of apps created specifically for the project.
        """
        return []

    def get_third_party_apps(self):
        """
        Return a list of third party dependencies.
        """
        return []

    def get_django_contrib_apps(self):
        """
        Return a list of apps from django.contrib.
        """

        apps = []
        if self.USE_DJANGO_ADMIN:
            apps.append('django.contrib.admin')
        if self.USE_DJANGO_USERS:
            apps.extend(['django.contrib.auth', 'django.contrib.sessions'])
        apps.extend(['django.contrib.contenttypes',
                     'django.contrib.messages',
                     'django.contrib.sites'])
        if self.SERVE_STATIC_FILES:
            apps.append('django.contrib.staticfiles')
        return apps

    def with_app(self, app, app_list, deps=None, optdeps=None):
        """
        Insert app on list, respecting its dependencies.

        Usage:
            Useful for third parties declare dependency classes:

            class MyAppConf(InstalledAppsConf):
                def get_third_party_apps(self):
                    apps = super().get_third_party_apps()
                    return self.with_app('my_app', apps, deps=['dep1', 'dep2'])

        Args:
            app:
                App name.
            app_list:
                List of apps.
            deps:
                List of hard dependencies. All apps in that list are also
                inserted in the result just after the inserted app.
            optdeps:
                List of optional dependencies. App will be inserted just
                before all apps in list if they appear in the app list.

        Returns:
            A list of apps.
        """
        apps = [app]
        apps.extend(x for x in (deps or ()) if x not in app_list)
        all_deps = set(deps or ())
        all_deps.update(optdeps or ())

        result = []
        for app in app_list:
            if apps and app in all_deps:
                result.extend(apps)
                apps = None
            result.append(app)
        return result

    @staticmethod
    def filter_installed_apps(apps):
        """
        Return only the installed apps in the following list.

        Each item may be either a string with the app name or a tuple/list with
        the app name followed by all of its dependencies.
        """
        filtered = []
        is_installed = (lambda dep: importlib.util.find_spec(dep) is not None)

        for app in apps:
            if isinstance(app, (tuple, list)):
                app, *deps = app
            else:
                app = deps = app
            if all(is_installed(dep) for dep in deps):
                filtered.append(app)

        return filtered



from .environment import EnvironmentConf
from ..tools import secret_hash


class SecurityConf(EnvironmentConf):
    """
    Security options.
    """

    def finalize(self, settings):
        settings = super().finalize(settings)
        if not settings.get('SECRET_KEY'):
            settings['SECRET_KEY'] = secret_hash(settings)
        return settings

    def get_secret_key(self):
        """
        WARNING: keep the secret key used in production secret! We generate a
        secret from a hash of the current settings during the .finalize() phase.
        this is ok for local development, but may be insecure/inconvenient for
        """
        value = self.env.str('DJANGO_SECRET_KEY', default=None)
        if not value:
            if self.ENVIRONMENT in ('local', 'test'):
                return self.ENVIRONMENT
            else:
                return None
        return value

    # https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators
    def get_auth_password_validators(self):
        """
        Password validation
        """
        prefix = 'django.contrib.auth.password_validation'
        validators = [
            'UserAttributeSimilarityValidator', 'MinimumLengthValidator',
            'CommonPasswordValidator', 'NumericPasswordValidator',
        ]
        return [{'NAME': f'{prefix}.{x}'} for x in validators]

    def get_allowed_hosts(self):
        return self.env('DJANGO_ALLOWED_HOSTS', type=list, default=['localhost'])

    def get_password_hashers(self):
        if self.ENVIRONMENT == 'testing':
            return [
                'django.contrib.auth.hashers.MD5PasswordHasher',
            ]
        return [
            'django.contrib.auth.hashers.Argon2PasswordHasher',
            'django.contrib.auth.hashers.PBKDF2PasswordHasher',
            'django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher',
            'django.contrib.auth.hashers.BCryptSHA256PasswordHasher',
            'django.contrib.auth.hashers.BCryptPasswordHasher',
        ]



from django.test import Client

LOGIN_REGEX = r'^/login/*.$'


class UrlChecker:
    """
    Base test class that checks the response code for specific urls in an app.
    """

    def __init__(self, urls=None, posts=None, login_regex=LOGIN_REGEX, client=None):
        self.client = client or Client()
        self.urls = urls or {}
        self.posts = posts or {}
        self.login_regex = login_regex

    #
    # Get errors from a URL list
    #
    def find_errors(self, url, code):
        """
        Check if client responds to given url with the provided status code.
        """
        response = self.client.get(url)
        if response.status_code not in code:
            return {url: response}
        else:
            return {}

    def get_url_codes(self, url, default):
        """
        Normalize an url which can be of form "string" or ("string", *codes)
        and return a pair of url string and the corresponding codes.
        """
        if isinstance(url, str):
            return url, default
        else:
            url, codes = url
            return url, codes

    def get_url_errors(self, urls, default_codes) -> dict:
        """
        Return a mapping with errors without changing the client state.
        """
        errors = {}
        for url in urls:
            url, codes = self.get_url_codes(url, default_codes)
            errors.update(self.find_errors(url, codes))
        return errors

    def check_public_urls(self, urls) -> dict:
        """
        Return a mapping of url to their corresponding errors when an error
        occurs.
        """
        self.client.logout()
        return self.get_url_errors(urls, (200, 301, 302))

    def check_user_access(self, urls, user) -> dict:
        """
        Return a mapping of url to their corresponding errors when an error
        occurs.
        """
        self.client.force_login(user)
        return self.get_url_errors(urls, (200, 301, 302))

    def check_login_required(self, urls) -> dict:
        """
        Return a mapping of url to their corresponding errors for failed login
        redirects.
        """
        self.client.logout()
        url_map = dict(urls)
        url_map.pop(None, None)
        urls = []
        for url_list in url_map.values():
            urls.extend(url_list)
        return self.get_url_errors(urls, (302, 404))

    def check_restricted_pages(self, urls, user) -> dict:
        """
        Return a mapping of url to their corresponding errors for cases that
        grant unwanted access to resources.
        """
        self.client.force_login(user)
        return self.get_url_errors(urls, (302, 404))

    def check_url_errors(self, users: dict) -> dict:
        """
        Runs all checks and create a dictionary with all errors.
        """
        urls = dict(self.urls)
        errors = {}

        # Public urls
        errors.update(self.check_public_urls(urls.pop(None, ())))

        # Test for each user
        for name, url_list in urls.items():
            user = users[name]
            errors.update(self.check_user_access(url_list, user))

        # Test login redirects
        errors.update(self.check_login_required(urls))

        # Test permissions
        if 'user' in users:
            user = users['user']
            url_list = []
            for user, user_list in users.items():
                if user not in (None, 'user'):
                    url_list.append(url_list)
            if url_list:
                errors.update(self.check_restricted_pages(url_list, user))

        return errors



from django.conf import settings as _settings

BOOGIE_REST_API_SCHEME = getattr(_settings, 'BOOGIE_REST_API_SCHEME', None)


def get_scheme(request):
    return BOOGIE_REST_API_SCHEME or request.scheme


def get_url_prefix(request):
    if request is None:
        return ''
    return f'{get_scheme(request)}://{request.get_host()}'



from functools import lru_cache

import pytest
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.http import Http404
from django.urls import resolve

from .client import Client
from .crawler import check_link_errors
from .urlchecker import LOGIN_REGEX, UrlChecker as _UrlChecker
from ..utils.text import snake_case

User = get_user_model()


#
# URL Tester
#
class URLTesterMeta(type):
    """
    Metaclass for UrlTester.
    """

    def __new__(mcs, name, bases, namespace, base=False):
        not_none = (lambda x: x is not None)
        users = tuple(sorted({
            *filter(not_none, namespace.get('urls', {}).keys()),
            *filter(not_none, namespace.get('posts', {}).keys()),
        }))
        if not base:
            namespace['test_urls'] = make_test_urls(users)
        return type.__new__(mcs, name, bases, namespace)


class UrlTester(metaclass=URLTesterMeta, base=True):
    """
    Base class for tests.
    """

    @pytest.fixture
    def data(self):
        return None

    @pytest.fixture
    def user(self, db):
        return self.make_user('user', email='user@user.com')

    @pytest.fixture
    def author(self, db):
        return self.make_user('author', email='author@author.com')

    @pytest.fixture
    def admin(self, db):
        return self.make_user('admin', email='admin@admin.com',
                              is_superuser=True, is_staff=True)

    def make_user(self, *args, **kwargs):
        return User.objects.create_user(*args, **kwargs)

    urls = {}
    posts = {}
    login_regex = LOGIN_REGEX


#
# Web crawling
#
class CrawlerTester:
    """
    Base class for tests based on crawling your website from a initial URL.

    Can be configured by overriding the following attributes:

        bases (list of urls):
            List of (url, user) pairs used as the starting points of web
            crawling.
        skip (list of urls):
            URLs that should be skipped during testing. The crawler will never
            visit those URLs unless they are a starting url in bases.
        xfail (list of urls):
            URLs that are expected to have failures.
        must_visit (list of urls):
            Checks if all URLs in that list were visited after crawling ends.
        user (str):
            Name of user instance fixture.
    """

    root = '/'
    user = 'user'
    skip = ()
    xfail = ()
    must_visit = ()
    log = print
    error_class = AssertionError

    @pytest.fixture
    def data(self):
        return {None: None}

    @pytest.fixture
    def conf(self, request):
        return {'user': request.getfixturevalue('user')}

    @pytest.mark.django_db
    def test_reaches_all_urls(self, data, conf):
        """
        Test if it fails in some view when crawling from a starting URL.
        """
        errors = {}
        user = conf['user']
        root = self.root
        if isinstance(root, str):
            root = [root]

        for url in root:
            try:
                check_link_errors(
                    url, visit=self.must_visit, skip=self.skip,
                    errors=self.xfail, user=user, log=self.log,
                )
            except AssertionError as ex:
                errors.update(ex.args[0])
        if errors:
            raise self.error_class(errors)


#
# Testing models
#
class ModelTester:
    """
    A helper class that makes it easier for testing models.
    """
    model: type
    representation: str
    fixture_name: str
    absolute_url: str
    db = False

    @pytest.fixture
    def instance(self, request):
        if not hasattr(self, 'fixture_name'):
            name = snake_case(self.model.__name__)
        else:
            name = self.fixture_name
        if self.db:
            request.getfixturevalue('db')
        return request.getfixturevalue(name)

    def get_examples(self):
        """
        Return a dictionary with some exemplary models with expected
        characteristics. Sub-classes can define any one of the entries bellow::

        - clean: passes a .full_clean() check
        - invalid: raises ValidationError during .full_clean()
        """
        return {}

    def test_django_model_interface(self, instance):
        """
        Test if model implements a basic Django Model interface correctly.
        """
        model = self.model
        self.check_improperly_configured(instance)

        # Check representation
        assert str(instance) == self.representation

        # Check absolute url
        if hasattr(model, 'get_absolute_url'):
            assert instance.get_absolute_url() == self.absolute_url
            try:
                resolve(self.absolute_url)
            except Http404:
                raise AssertionError('absolute_url resulted on a 404')

        # Run examples
        for name, example in self.get_examples().items():
            if name == 'clean':
                example.full_clean()
            elif example == 'invalid':
                with pytest.raises(ValidationError):
                    example.full_clean()
            else:
                raise ImproperlyConfigured(f'invalid example: {name}')

    def check_improperly_configured(self, instance):
        """
        Check if test class was correctly set up for instance.
        """
        model = self.model
        requires_url = hasattr(model, 'get_absolute_url')

        # Check success
        if (hasattr(self, 'representation')
                and (hasattr(self, 'absolute_url') or not requires_url)):
            return

        print('HINT: You can create a valid tester class by replacing it by\n'
              'the following code:')
        print(make_test_class(self, instance))
        print('\nPlease revise to see if inferred properties are correct.')
        raise ImproperlyConfigured('Model tester class is not configured')


#
# Fixtures
#
@pytest.fixture
def client(db):
    """Standard test client"""
    return Client()


@pytest.fixture
def user_client(db, user):
    """Client logged in as 'user'"""
    client = Client()
    client.force_login(user)
    return client


@pytest.fixture
def admin_client(db, admin):
    """Client logged in as 'admin'"""
    client = Client()
    client.force_login(admin)
    assert admin.is_staff, f'User {admin} does not have staff privileges'
    return client


@pytest.fixture
def author_client(db, author):
    """Client logged in as 'author'"""
    client = Client()
    client.force_login(author)
    return client


#
# Utilities
#
def make_test_class(model_tester, instance):
    """
    Return a string with a plausible source code representation for a
    ModelTester test class.
    """

    base_names = ', '.join(cls.__name__ for cls in type(model_tester).__bases__)
    base = f"""
    class {model_tester.__class__.__name__}({base_names}):
        model = models.{instance.__class__.__name__}
        representation = {repr(str(instance))}"""
    if model_tester.db:
        base += '\n        db = True'
    if hasattr(instance, 'get_absolute_url'):
        base += '\n        absolute_url = {instance.get_absolute_url()!r}'
    return base


@lru_cache(32)
def make_test_urls(users):
    """
    Create the test_urls() function for the given list of users.

    Users are passed to function as fixtures.
    """
    inputs = ', '.join(['data', 'db', 'client', *users])
    src = (
        f'def test_urls(self, {inputs}):\n'
        f'    check_urls_from_locals(locals())'
    )

    # Eval code and fetch result
    ns = {}
    exec(src, {'check_urls_from_locals': check_urls_from_locals}, ns)
    return ns['test_urls']


def check_urls_from_locals(variables):
    """
    Implements the logic used by the dynamic test_urls class.
    """
    self = variables.pop('self')
    variables.pop('data')
    variables.pop('db')
    client = variables.pop('client')
    checker = _UrlChecker(self.urls, self.posts,
                          self.login_regex, client=client)
    errors = checker.check_url_errors(users=variables)
    if errors:
        for url, value in errors.items():
            code = value.status_code
            print(f'Error fetching {url}: got status code {code}')
        raise AssertionError(f'errors found at given urls: {list(errors)}')



import inspect
import operator as op
from functools import partial
from functools import update_wrapper

from rules.predicates import Context, NO_VALUE, Predicate, _context

from sidekick.operators import SYMBOLS


def binary(op, reverse=False, symbol=None):
    symbol = symbol or SYMBOLS[op]

    def binary_op(self, other):
        if isinstance(other, Value):
            fn = (lambda *args: self._combine_binop(other, op, args))
            name = f'{self.name} {symbol} {other.name}'
            return Value(fn, name)
        else:
            transform = lambda x: op(x, other)
            fn = (lambda *args: self._transform(transform, args))
            name = f'{self.name} {symbol} {repr(other)}'
            return Value(fn, name)

    if reverse:
        def rbinary_op(self, other):
            fn = (lambda *args: self._transform(partial(op, other), args))
            name = f'{self.name} {symbol} {repr(other)}'
            return Value(fn, name)

        return rbinary_op

    return binary_op


def unary(op, symbol):
    symbol = symbol or SYMBOLS.get(op, op.__name__)

    def unary(self):
        fn = (lambda *args: self._transform(op, args))
        name = f'{symbol} {self.name}'
        return Value(fn, name)

    return unary


class Value:
    def __init__(self, fn, name=None, bind=False):  # noqa: C901
        # fn can be a callable with any of the following signatures:
        #   - fn(obj=None, user=None)
        #   - fn(obj=None)
        #   - fn()
        if isinstance(fn, Value):
            fn, num_args, var_args, name = fn.fn, fn.num_args, fn.var_args, name or fn.name
        elif isinstance(fn, partial):
            argspec = inspect.getfullargspec(fn.func)
            var_args = argspec.varargs is not None
            num_args = len(argspec.args) - len(fn.args)
            if inspect.ismethod(fn.func):
                num_args -= 1  # skip `self`
            name = fn.func.__name__
        elif inspect.ismethod(fn):
            argspec = inspect.getfullargspec(fn)
            var_args = argspec.varargs is not None
            num_args = len(argspec.args) - 1  # skip `self`
        elif inspect.isfunction(fn):
            argspec = inspect.getfullargspec(fn)
            var_args = argspec.varargs is not None
            num_args = len(argspec.args)
        elif callable(fn):
            callfn = getattr(fn, '__call__')
            argspec = inspect.getfullargspec(callfn)
            var_args = argspec.varargs is not None
            num_args = len(argspec.args) - 1  # skip `self`
            name = name or type(fn).__name__
        else:
            raise TypeError('Incompatible value function.')
        if bind:
            num_args -= 1
        if num_args > 2:
            raise TypeError('Value function must receive at most 2 arguments')
        self.fn = fn
        self.num_args = num_args
        self.var_args = var_args
        self.name = name or fn.__name__
        self.bind = bind

    def __repr__(self):
        type_name = type(self).__name__
        return '<%s:%s object at %s>' % (type_name, self, hex(id(self)))

    # Value is not a subclass of Predicate, but shares many functionality
    # We simply copy the common methods here.
    __str__ = Predicate.__str__
    __call__ = Predicate.__call__
    context = Predicate.context
    skip = Predicate.skip

    # Binary operators
    __add__ = binary(op.add)
    __radd__ = binary(op.add, reverse=True)
    __sub__ = binary(op.sub)
    __rsub__ = binary(op.sub, reverse=True)
    __mul__ = binary(op.mul)
    __rmul__ = binary(op.mul, reverse=True)
    __truediv__ = binary(op.truediv)
    __rtruediv__ = binary(op.truediv, reverse=True)

    def compute(self, obj=NO_VALUE, user=NO_VALUE):
        """
        The canonical method to invoke values.
        """
        args = tuple(arg for arg in (obj, user) if arg is not NO_VALUE)
        _context.stack.append(Context(args))
        try:
            return self._apply(*args)
        finally:
            _context.stack.pop()

    def _apply(self, *args):
        # Internal method that is used to invoke the predicate with the
        # proper number of positional arguments, inside the current
        # invocation context.
        if self.var_args:
            callargs = args
        elif self.num_args > len(args):
            callargs = args + (None,) * (self.num_args - len(args))
        else:
            callargs = args[:self.num_args]
        if self.bind:
            callargs = (self,) + callargs
        return self.fn(*callargs)

    def _transform(self, func, args):
        result = self._apply(*args)
        return func(result)

    def _combine_binop(self, other, op, args):
        left = self._apply(*args)
        right = other._apply(*args)
        return op(left, right)


def value(fn=None, name=None, **options):
    """
    Decorator that constructs a ``Value`` instance from any function::

    >>> @value
    ... def status_display(question):
    ...     if question.status == question.OPEN:
    ...         return 'Receiving submissions'
    ...     else:
    ...         return 'Closed'
    """
    if not name and not callable(fn):
        name = fn
        fn = None

    def inner(fn):
        if isinstance(fn, Value):
            return fn
        p = Value(fn, name, **options)
        update_wrapper(p, fn)
        return p

    if fn:
        return inner(fn)
    else:
        return inner



from sidekick import import_later

# Function importer
_function = (lambda x: import_later(x, package=__package__))

# Lazy imports of checker functions
check_link_errors = _function('.crawler:check_link_errors')
factory = _function('.factories:factory')
